/***************************************************************************
 *   Copyright (C) 1999-2002 by Bernd Gehrmann                             *
 *   bernd@mail.berlios.de                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License, version 2       *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "contentswidget.h"
#include "global.h"

#include <stdlib.h>
#include <qapplication.h>
#include <qcursor.h>
#include <qdir.h>
#include <qdom.h>
#include <qfile.h>
#include <qfileinfo.h>
#include <qheader.h>
#include <qlistbox.h>
#include <qlistview.h>
#include <qpair.h>
#include <qpixmap.h>
#include <qpopupmenu.h>
#include <qregexp.h>
#include <qsettings.h>
#include <qstringlist.h>
#include <qtabwidget.h>
#include <qvaluelist.h>


class TreeItem : public QListViewItem
{
public:
    TreeItem(QListView *lv, const QString &text, const QString &fileName)
        : QListViewItem(lv, text), fn(fileName)
    {}
    TreeItem(QListViewItem *parent, const QString &text, const QString &fileName)
        : QListViewItem(parent, text), fn(fileName)
    {}

    QString url() const
    { return fn; }

private:
    QString fn;
};

        
class FolderItem : public QListBoxText
{
public:
    FolderItem(QListBox *lb)
        : QListBoxText(lb)
    {}
    virtual void activate(QListView *lv) = 0;
};


/*************************************/
/* Folders from .tocs files          */
/*************************************/

class TocItem : public FolderItem
{
public:
    TocItem(QListBox *lb, const QString &fileName)
        : FolderItem(lb), fn(fileName), loaded(false)
    {
        QString text = QFileInfo(fileName).baseName();
        text.replace(QRegExp(_("_")), _(" "));
        setText(text);
    }

    virtual void activate(QListView *lv)
    {
        lv->setSorting(-1);

        if (!loaded)
            load();
        
        QDomElement docEl = dom.documentElement();
        QString base;
        QListViewItem *lastChildItem = 0;

        /**/
        //        TreeItem *item = new TreeItem(lv, "testqt", "/home/bernd/kde3src/qt3/doc/html/qobject.html");
        //        TreeItem *item = new TreeItem(lv, "testqt", "/home/bernd/kde3src/dox/dox/test.html");
        /**/
        QDomElement childEl = docEl.firstChild().toElement();
        while (!childEl.isNull()) {
            if (childEl.tagName() == _("tocsect1")) {
                QString name = childEl.attribute(_("name"));
                QString url = childEl.attribute(_("url"));
                QString fullurl = url.isEmpty()? url : (base + url);
                TreeItem *item = new TreeItem(lv, name, fullurl);
                if (lastChildItem)
                    item->moveItem(lastChildItem);
                lastChildItem = item;
                parseDom(item, base, childEl, 2);
            } else if (childEl.tagName() == _("base")) {
                base = childEl.attribute(_("href"));
                if (!base.isEmpty())
                    base += _("/");
            }
            childEl = childEl.nextSibling().toElement();
        }
        
    }

    void parseDom(QListViewItem *parent, const QString &base,
                  QDomElement &el, int level)
    {
        QString tag = _("tocsect%1").arg(level);
        QListViewItem *lastChildItem = 0;
        QDomElement childEl = el.firstChild().toElement();
        while (!childEl.isNull()) {
            if (childEl.tagName() == tag) {
                QString name = childEl.attribute(_("name"));
                QString url = childEl.attribute(_("url"));
                QString fullurl = url.isEmpty()? url : (base+url);
                TreeItem *item = new TreeItem(parent, name, fullurl);
                if (lastChildItem)
                    item->moveItem(lastChildItem);
                lastChildItem = item;
                parseDom(item, base, childEl, level+1);
            }
            childEl = childEl.nextSibling().toElement();
        }
    }

    virtual void load()
    {
        loaded = true;
        QFile f(fn);
        if (!f.open(IO_ReadOnly)) {
            qDebug("Could not read doc toc: %s", fn.latin1());
            return;
        }
        if (!dom.setContent(&f) || dom.doctype().name() != _("doxtoc0")) {
            qDebug("Not a valid doxtoc file: %s", fn.latin1());
            return;
        }
        f.close();
    }
    
private:
    QString fn;
    QDomDocument dom;
    bool loaded;
};


/*************************************/
/* Folder "Documentation Base"       */
/*************************************/

class DocbaseItem : public FolderItem
{
public:
    DocbaseItem(QListBox *lb)
        : FolderItem(lb)
    {
        setText(ContentsWidget::trUtf8("Docbase"));
    }

    virtual void activate(QListView *lv)
    {
        lv->setSorting(0);

        if (entries.isEmpty())
            readDocbaseDir();
        
        QValueList<StringPair>::ConstIterator it;
        for (it = entries.begin(); it != entries.end(); ++it)
            (void) new TreeItem(lv, (*it).first, (*it).second);
    }

    void readDocbaseDir()
    {
        QDir d(_("/usr/share/doc-base"));
        QStringList fileList = d.entryList(_("*"), QDir::Files);
        QStringList::Iterator it;
        for (it = fileList.begin(); it != fileList.end(); ++it) {
            FILE *f = fopen(QFile::encodeName(d.filePath(*it)), "r");
            if (f != 0) {
                readDocbaseFile(f);
                fclose(f);
            }
        }
    }
    
    void readDocbaseFile(FILE *f)
    {
        char buf[1024];
        QCString title;
        bool html = false;
        while (fgets(buf, sizeof buf, f)) {
            QCString s = buf;
            if (s.right(1) == "\n")
                s.truncate(s.length()-1); // chop
            
            if (s.left(7) == "Title: ")
                title = s.mid(7, s.length()-7);
            else if (s.left(8) == "Format: ")
                html = s.find("HTML", 8, false) != -1;
            else if (s.left(7) == "Index: "
                     && html && !title.isEmpty()) {
                QString filename = _(s.mid(7, s.length()-7));
                entries.append(StringPair(_(title), filename));
                break;
            }
            else if (s.left(9) == "Section: "
                     && s.find("programming", 9, false) == -1)
                break;
        }
    }
    
private:
    typedef QPair<QString,QString> StringPair;
    QValueList<StringPair> entries;
};


/*************************************/
/* Folder "Man pages"                */
/*************************************/

class ManpageItem : public FolderItem
{
public:
    ManpageItem(QListBox *lb)
        : FolderItem(lb),
          whatisRegExp(_("^(\\S+)\\s+\\(([^)]+)\\)\\s*(\\[[^\\]]+\\])?\\s*-\\s+(.*)$"))
    {
        setText(ContentsWidget::trUtf8("Man pages"));

        niceSectionNames.insert(_("1"),        ContentsWidget::trUtf8("Commands (1)"));
        niceSectionNames.insert(_("1m"),       ContentsWidget::trUtf8("Commands (1m)"));
        niceSectionNames.insert(_("1x"),       ContentsWidget::trUtf8("Commands (1x)"));
        niceSectionNames.insert(_("2"),        ContentsWidget::trUtf8("System calls (2)"));
        niceSectionNames.insert(_("3"),        ContentsWidget::trUtf8("Library calls (3)"));
        niceSectionNames.insert(_("3curses"),  ContentsWidget::trUtf8("Curses Library (3curses)"));
        niceSectionNames.insert(_("3form"),    ContentsWidget::trUtf8("Library calls (3form)"));
        niceSectionNames.insert(_("3gle"),     ContentsWidget::trUtf8("GLE Library (3gle)"));
        niceSectionNames.insert(_("3menu"),    ContentsWidget::trUtf8("Library calls (3menu)"));
        niceSectionNames.insert(_("3ncurses"), ContentsWidget::trUtf8("Library calls (3ncurses)"));
        niceSectionNames.insert(_("3pm"),      ContentsWidget::trUtf8("Perl modules (3pm)"));
        niceSectionNames.insert(_("3qt"),      ContentsWidget::trUtf8("Qt Library calls (3qt)"));
        niceSectionNames.insert(_("3t"),       ContentsWidget::trUtf8("TIFF Library calls (3t)"));
        niceSectionNames.insert(_("3thr"),     ContentsWidget::trUtf8("pthreads Library calls (3thr)"));
        niceSectionNames.insert(_("3x"),       ContentsWidget::trUtf8("X11 Library calls (3x)"));
        niceSectionNames.insert(_("4"),        ContentsWidget::trUtf8("Special files (4)"));
        niceSectionNames.insert(_("5"),        ContentsWidget::trUtf8("File formats and conventions (5)"));
        niceSectionNames.insert(_("5x"),       ContentsWidget::trUtf8("File formats and conventions (5x)"));
        niceSectionNames.insert(_("6"),        ContentsWidget::trUtf8("Games (6)"));
        niceSectionNames.insert(_("7"),        ContentsWidget::trUtf8("Macro packages and conventions (7)"));
        niceSectionNames.insert(_("8"),        ContentsWidget::trUtf8("System management commands (8)"));
        niceSectionNames.insert(_("9"),        ContentsWidget::trUtf8("Kernel routines (9)"));
        niceSectionNames.insert(_("n"),        ContentsWidget::trUtf8("Tcl/Tk"));
    }

    virtual void activate(QListView *lv)
    {
        lv->setSorting(0);
        
        if (sections.isEmpty())
            readManpages();

        SectionMap::ConstIterator sectit;
        for (sectit = sections.begin(); sectit != sections.end(); ++sectit) {
            QString sectName = sectit.key();
            QString niceSectName = niceSectionNames[sectName];
            if (niceSectName.isNull())
                niceSectName = sectName;
            QListViewItem *sectItem = new TreeItem(lv, niceSectName, QString::null);
            
            QValueList<StringPair> pairList = sectit.data();
            QValueList<StringPair>::ConstIterator it;
            for (it = pairList.begin(); it != pairList.end(); ++it)
                (void) new TreeItem(sectItem, (*it).first, (*it).second);
        }
    }

    void readManpages()
    {
        qApp->setOverrideCursor(Qt::waitCursor);
    
        QStringList paths = QStringList::split(_(":"), _(getenv("MANPATH")));
        QStringList::ConstIterator it;
        for (it = paths.begin(); it != paths.end(); ++it) {
            QString whatis = (*it) + _("/whatis");
            if (0 /*QFile::exists(whatis)*/) {
                QFile f(whatis);
                if (f.open(IO_ReadOnly)) {
                    QTextStream *stream = new QTextStream(&f);
                    readWhatisFile(stream);
                    delete stream;
                    f.close();
                }
            } else {
#if HAVE_SANEWHATIS
                QCString command = "whatis -M ";
                command += QFile::encodeName(*it);
                command += " -w '*' < /dev/null";
                FILE *f = ::popen(command.data(), "r");
                if (f) {
                    QTextStream *stream = new QTextStream(f, IO_ReadOnly);
                    readWhatisFile(stream);
                    delete stream;
                    ::pclose(f);
                }
#else
                readManDir(*it);
#endif
            }
        }

        qApp->restoreOverrideCursor();
    }
    
    void readWhatisFile(QTextStream *stream)
    {
        while (!stream->atEnd()) {
            QString line = stream->readLine();
            if (whatisRegExp.exactMatch(line)) {
                QString page = whatisRegExp.cap(1);
                QString section = whatisRegExp.cap(2);
                QString title = whatisRegExp.cap(4);
                StringPair pair(_("%1 - %2").arg(page).arg(title),
                                _("man://%1.%2").arg(page).arg(section));
                SectionMap::Iterator sectit = sections.find(section);
                if (sectit != sections.end()) {
                    (*sectit).append(pair);
                } else {
                    QValueList<StringPair> pairList;
                    pairList.append(pair);
                    sections.insert(section, pairList);
                }
            } else {
                qDebug("Unmatched line %s", line.latin1());
            }
        }
    }

    void readManDir(const QString &dirName)
    {
        QDir d(dirName);
        QStringList dirs = d.entryList(_("man*"), QDir::Dirs);
        QStringList::ConstIterator dit;
        for (dit = dirs.begin(); dit != dirs.end(); ++dit) {
            qDebug("Listing %s", (*dit).local8Bit().data());
            QDir d2(dirName + _("/") + (*dit));
            QStringList files = d2.entryList(QDir::Files);
            QStringList::ConstIterator fit;
            for (fit = files.begin(); fit != files.end(); ++fit) {
                int pos2 = (*fit).findRev('.');
                if (pos2 < 3)
                    continue;
                int pos1 = (*fit).findRev('.', pos2-1);
                if (pos1 < 1)
                    continue;
                QString page = (*fit).left(pos1);
                QString section = (*fit).mid(pos1+1, pos2-pos1-1);
                StringPair pair(page, _("man://%1.%2").arg(page).arg(section));
                SectionMap::Iterator sectit = sections.find(section);
                if (sectit != sections.end()) {
                    (*sectit).append(pair);
                } else {
                    QValueList<StringPair> pairList;
                    pairList.append(pair);
                    sections.insert(section, pairList);
                }
            }
        }
    }

private:
    typedef QPair<QString,QString> StringPair;
    typedef QMap< QString,QValueList<StringPair> > SectionMap;
    SectionMap sections;
    QRegExp whatisRegExp;
    QMap<QString,QString> niceSectionNames;
};


/*************************************/
/* Folder "Info pages"                */
/*************************************/

class InfopageItem : public FolderItem
{
public:
    InfopageItem(QListBox *lb)
        : FolderItem(lb),
          sectionRegExp(_("^[A-Za-z].*")),
          itemRegExp(_("^\\* ([^:]+):(:| \\(([^)]+)\\)(([^.]*)\\.|\\s)).*"))
    {
        setText(ContentsWidget::trUtf8("Info pages"));
    }

    virtual void activate(QListView *lv)
    {
        lv->setSorting(0);

        if (sections.isEmpty())
            readInfoPages();

        SectionMap::ConstIterator sectit;
        for (sectit = sections.begin(); sectit != sections.end(); ++sectit) {
            QListViewItem *sectItem = new TreeItem(lv, sectit.key(), QString::null);
            QValueList<StringPair> pairList = sectit.data();
            QValueList<StringPair>::ConstIterator it;
            for (it = pairList.begin(); it != pairList.end(); ++it)
                (void) new TreeItem(sectItem, (*it).first, (*it).second);
        }
    }
                     
    void readInfoPages()
    {
        QStringList paths = QStringList::split(_(":"), _(getenv("INFOPATH")));
        QStringList::ConstIterator it;
        for (it = paths.begin(); it != paths.end(); ++it) {
            QFile f((*it) + _("/dir"));
            if (f.open(IO_ReadOnly)) {
                readDirFile(&f);
                f.close();
            }
        }
    }
    
    void readDirFile(QIODevice *f)
    {
        QTextStream stream(f);
        QString line;
        do {
            line = stream.readLine();
            if (line.isNull())
                return;
        } while (!line.startsWith(_("* Menu:")));

        QString section = ContentsWidget::trUtf8("General");

        do {
            line = stream.readLine();
            if (line.isNull())
                return;

            if (sectionRegExp.exactMatch(line)) {
                section = line;
            } else if (itemRegExp.exactMatch(line)) {
                QString title = itemRegExp.cap(1);
                QString page = QFileInfo(itemRegExp.cap(3)).fileName();
                if (page.isEmpty())
                    page = title.lower();
                QString node = itemRegExp.cap(5);
                if (!node.isEmpty())
                    node.prepend(_("/"));
                StringPair pair(title, _("info://%1%2").arg(page).arg(node));
                SectionMap::Iterator sectit = sections.find(section);
                if (sectit != sections.end()) {
                    (*sectit).append(pair);
                } else {
                    QValueList<StringPair> pairList;
                    pairList.append(pair);
                    sections.insert(section, pairList);
                }
            }
        } while(1);
    }
    
private:
    typedef QPair<QString,QString> StringPair;
    typedef QMap< QString,QValueList<StringPair> > SectionMap;
    SectionMap sections;
    QRegExp sectionRegExp, itemRegExp;
};


/*************************************/
/* Folder "KDE manuals"              */
/*************************************/

class KDEItem : public FolderItem
{
public:
    KDEItem(QListBox *lb)
        : FolderItem(lb)
    {
        setText(ContentsWidget::trUtf8("KDE manuals"));
    }

    virtual void activate(QListView /* *lv */)
    {
    }
};


/*****************************/
/* The ContentsWidget itself */
/*****************************/


ContentsWidget::ContentsWidget(QTabWidget *parent, const char *name)
    : QSplitter(Vertical, parent, name)
{
    tabParent = parent;

    overview = new QListBox(this, "overview");
    overview->setFrameStyle(Panel | Sunken);

    details = new QListView(this, "details");
    details->setResizeMode(QListView::LastColumn);
    details->setFrameStyle(Panel | Sunken);
    details->setRootIsDecorated(true);
    details->setSorting(-1);
    details->header()->hide();
    details->addColumn(_(""));

    setOpaqueResize(true);
    setFocusProxy(overview);
    
    QSettings *settings = Global::settings();
    int size1 = settings->readNumEntry(_("/Dox/ContentsWidget/SplitterSize1"), 100);
    int size2 = settings->readNumEntry(_("/Dox/ContentsWidget/SplitterSize2"), -1);
    if (size2 != -1) {
        QValueList<int> sizes;
        sizes << size1 << size2;
        setSizes(sizes);
    }

    // Debian only
    if (QFile::exists(_("/usr/share/doc-base")))
        (void) new DocbaseItem(overview);
    // Man pages
    (void) new ManpageItem(overview);
    (void) new InfopageItem(overview);

    QStringList tocs = Global::resources(_("share/dox/tocs"));
    tocs.sort();
    
    QStringList::Iterator tit;
    for (tit = tocs.begin(); tit != tocs.end(); ++tit)
        (void) new TocItem(overview, *tit);

    connect( overview, SIGNAL(returnPressed(QListBoxItem*)),
             this, SLOT(overviewClicked(QListBoxItem*)) );
    connect( overview, SIGNAL(clicked(QListBoxItem*)),
             this, SLOT(overviewClicked(QListBoxItem*)) );
    connect( overview, SIGNAL(onItem(QListBoxItem*)),
             this, SLOT(onOverviewItem(QListBoxItem*)) );
    connect( overview, SIGNAL(onViewport()),
             this, SLOT(onOverviewViewport()) );
    connect( details, SIGNAL(returnPressed(QListViewItem*)),
             this, SLOT(detailsClicked(QListViewItem*)) );
    connect( details, SIGNAL(clicked(QListViewItem*)),
             this, SLOT(detailsClicked(QListViewItem*)) );
    connect( details, SIGNAL(contextMenuRequested(QListViewItem*, const QPoint &, int)),
             this, SLOT(detailsContextMenuRequested()) );
    connect( details, SIGNAL(onItem(QListViewItem*)),
             this, SLOT(onDetailsItem(QListViewItem*)) );
    connect( details, SIGNAL(onViewport()),
             this, SLOT(onDetailsViewport()) );
}


ContentsWidget::~ContentsWidget()
{
    QValueList<int> sizepair = sizes();
    QSettings *settings = Global::settings();
    settings->writeEntry(_("/Dox/ContentsWidget/SplitterSize1"), sizepair[0]);
    settings->writeEntry(_("/Dox/ContentsWidget/SplitterSize2"), sizepair[1]);
}


void ContentsWidget::showEvent(QShowEvent *)
{
    overview->setFocus();
}


void ContentsWidget::overviewClicked(QListBoxItem *item)
{
    if (!item)
        return;

    // We assume here that ALL items in the list view
    // are DocTreeItem's
    FolderItem *dtitem = static_cast<FolderItem*>(item);

    details->clear();
    dtitem->activate(details);
    details->setFocus();
}


void ContentsWidget::onOverviewItem(QListBoxItem *)
{
    overview->viewport()->setCursor(pointingHandCursor);
}


void ContentsWidget::onOverviewViewport()
{
    overview->viewport()->setCursor(arrowCursor);
}


void ContentsWidget::detailsClicked(QListViewItem *item)
{
    if (!item)
        return;

    TreeItem *titem = static_cast<TreeItem*>(item);
    emit gotoFile(titem->url());
}


void ContentsWidget::detailsContextMenuRequested()
{
    QPopupMenu popup(this);
    int unfold = popup.insertItem(trUtf8("Unfold tree"));
    int fold = popup.insertItem(trUtf8("Fold tree"));
    int r = popup.exec(QCursor::pos());
    if (r != fold && r != unfold)
        return;

    QListViewItemIterator it(details);
    for (; it.current(); ++it)
        if (it.current()->firstChild())
            it.current()->setOpen(r == unfold);
}


void ContentsWidget::onDetailsItem(QListViewItem *item)
{
    if (!item)
        return;
    TreeItem *titem = static_cast<TreeItem*>(item);
    QString url = titem->url();

    emit highlighted(url);
    details->viewport()->setCursor(url.isEmpty()?
                                   arrowCursor : pointingHandCursor);
}


void ContentsWidget::onDetailsViewport()
{
    details->viewport()->setCursor(arrowCursor);
    emit highlighted(QString());
}


void ContentsWidget::showBook(const QString &name)
{
    tabParent->showPage(this);
    
    QListBoxItem *item = overview->findItem(name);
    if (item)
        overviewClicked(item);
}
