/***************************************************************************
 *   Copyright (C) 2001-2002 by Bernd Gehrmann                             *
 *   bernd@mail.berlios.de                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License, version 2       *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "indexwidget.h"
#include "global.h"

#include <qapplication.h>
#include <qfile.h>
#include <qfileinfo.h>
#include <qfontmetrics.h>
#include <qlabel.h>
#include <qlineedit.h>
#include <qlistbox.h>
#include <qprogressdialog.h>
#include <qregexp.h>
#include <qsettings.h>
#include <qstringlist.h>
#include <qtabwidget.h>
#include <qvbox.h>


class IndexFile
{
public:
    QString indexName;
    QString title;
    QString base;
};


class Topic
{
public:
    Topic()
    {}
    Topic(int fileno, const QString &title, const QString &relUrl)
        : file(fileno), text(title), url(relUrl)
    {}
    
    char file; // points to indexFiles list
    QString text;
    QString url;
};


class IndexItem : public QListBoxText
{
public:
    IndexItem(QListBox *lb, const QString &text)
        : QListBoxText(lb, text)
    {}
    void addTopic(int file, const QString &title, const QString &url)
    {
        topics.append(Topic(file, title, url));
    }

    QValueList<Topic> topics;
};


class TopicItem : public QListBoxText
{
public:
    TopicItem(QListBox *lb, const QString &text, const QString &fullUrl)
        : QListBoxText(lb, text), url(fullUrl)
    {}

    QString url;
};

        
IndexWidget::IndexWidget(QTabWidget *parent, const char *name)
    : QSplitter(Vertical, parent, name)
{
    tabParent = parent;

    QVBox *vbox = new QVBox(this);
    vbox->setSpacing(4);
    vbox->setMargin(4);
    
    QLabel *termLabel = new QLabel(trUtf8("Find &term:"), vbox);
    termEdit = new QLineEdit(vbox);

    indexListBox = new QListBox(vbox);
    indexListBox->setFrameStyle(Panel | Sunken);
    
    topicListBox = new QListBox(this);
    indexListBox->setFrameStyle(Panel | Sunken);

    setResizeMode(topicListBox, KeepSize);
    setOpaqueResize(true);
    setFocusProxy(termEdit);
    
    termLabel->setBuddy(termEdit);
    termEdit->installEventFilter(this);
    indexFiles.setAutoDelete(true);

    QSettings *settings = Global::settings();
    int size1 = settings->readNumEntry(_("/Dox/IndexWidget/SplitterSize1"), 100);
    int size2 = settings->readNumEntry(_("/Dox/IndexWidget/SplitterSize2"), -1);
    if (size2 != -1) {
        QValueList<int> sizes;
        sizes << size1 << size2;
        setSizes(sizes);
    }

    connect( termEdit, SIGNAL(textChanged(const QString &)),
             this, SLOT(textChanged(const QString &)) );
    connect( termEdit, SIGNAL(returnPressed()),
             this, SLOT(returnPressed()) );
    connect( indexListBox, SIGNAL(returnPressed(QListBoxItem*)),
             this, SLOT(indexItemSelected(QListBoxItem*)) );
    connect( indexListBox, SIGNAL(clicked(QListBoxItem*)),
             this, SLOT(indexItemSelected(QListBoxItem*)) );
    connect( indexListBox, SIGNAL(onItem(QListBoxItem*)),
             this, SLOT(onIndexItem(QListBoxItem*)) );
    connect( indexListBox, SIGNAL(onViewport()),
             this, SLOT(onIndexViewport()) );
    connect( topicListBox, SIGNAL(returnPressed(QListBoxItem*)),
             this, SLOT(topicItemSelected(QListBoxItem*)) );
    connect( topicListBox, SIGNAL(clicked(QListBoxItem*)),
             this, SLOT(topicItemSelected(QListBoxItem*)) );
    connect( topicListBox, SIGNAL(onItem(QListBoxItem*)),
             this, SLOT(onTopicItem(QListBoxItem*)) );
    connect( topicListBox, SIGNAL(onViewport()),
             this, SLOT(onTopicViewport()) );
}


IndexWidget::~IndexWidget()
{
    QValueList<int> sizepair = sizes();
    QSettings *settings = Global::settings();
    settings->writeEntry(_("/Dox/IndexWidget/SplitterSize1"), sizepair[0]);
    settings->writeEntry(_("/Dox/IndexWidget/SplitterSize2"), sizepair[1]);
}


void IndexWidget::showEvent(QShowEvent *e)
{
    QSplitter::showEvent(e);
    
    if (!indexFiles.isEmpty())
        return;

    QStringList books = Global::resources(_("share/dox/indices"));
    books.sort();

    qApp->setOverrideCursor(waitCursor);
    QProgressDialog *progress
        = new QProgressDialog(QString::null, QString::null, books.count(),
                              this, "progress", true);
    progress->setCaption(_("Dox"));
    progress->setCancelButton(0);
    QFontMetrics fm(progress->fontMetrics());
    int w = fm.width(trUtf8("Loading index file: %1").arg(_("ABCDEFGHIJKLMNOPQRSTUVW")));
    progress->setMinimumWidth(w);
    
    QStringList::Iterator bit;
    int i = 0;
    for (bit = books.begin(); bit != books.end(); ++bit) {
        QString text = QFileInfo(*bit).baseName();
        text.replace(QRegExp(_("_")), _(" "));
        progress->setLabelText(trUtf8("Loading index file: %1").arg(text));
        readIndexFromFile(*bit);
        progress->setProgress(++i);
        qApp->processEvents();
    }
    indexListBox->sort();

    delete progress;
    qApp->restoreOverrideCursor();
    termEdit->setFocus();
}


bool IndexWidget::eventFilter(QObject *o, QEvent *e)
{
    if (o == termEdit && e->type() == QEvent::KeyPress) {
        int key = static_cast<QKeyEvent*>(e)->key();
        if (key == Key_Up || key == Key_Down
            || key == Key_Prior || key == Key_Next) {
            QApplication::sendEvent(indexListBox, e);
            termEdit->setText(indexListBox->currentText());
            return true;
        }
    }

    return QSplitter::eventFilter(o, e);
}


void IndexWidget::textChanged(const QString &text)
{
    QListBoxItem *item
        = indexListBox->findItem(text, BeginsWith|CaseSensitive);
    if (item)
        indexListBox->setCurrentItem(item);
}


void IndexWidget::returnPressed()
{
    QListBoxItem *item = indexListBox->item(indexListBox->currentItem());
    if (!item)
        return;

    termEdit->setText(item->text());
    indexItemSelected(item);
}


void IndexWidget::indexItemSelected(QListBoxItem *item)
{
    if (!item)
        return;
    IndexItem *iitem = static_cast<IndexItem*>(item);
    
    topicListBox->clear();
    
    QValueList<Topic>::ConstIterator it;
    for (it = iitem->topics.begin(); it != iitem->topics.end(); ++it) {
        QString title = (*it).text;
        QString url = (*it).url;
        IndexFile *indexFile = indexFiles.at((*it).file);
        QString base = indexFile->base;
        QString bookTitle = indexFile->title;
        QString fullurl = base + url;
        
        new TopicItem(topicListBox,
                      trUtf8("%1 (%2)").arg(title).arg(bookTitle),
                      fullurl);
        if (iitem->topics.count() == 1)
            gotoFile(fullurl);
    }

    if (iitem->topics.count() > 1) {
        topicListBox->setFocus();
        topicListBox->setSelected(0, true);
    }
}


void IndexWidget::onIndexItem(QListBoxItem *)
{
    indexListBox->viewport()->setCursor(pointingHandCursor);
}


void IndexWidget::onIndexViewport()
{
    indexListBox->viewport()->setCursor(arrowCursor);
}


void IndexWidget::topicItemSelected(QListBoxItem *item)
{
    if (!item)
        return;
    TopicItem *titem = static_cast<TopicItem*>(item);

    emit gotoFile(titem->url);
}


void IndexWidget::onTopicItem(QListBoxItem *item)
{
    if (!item)
        return;
    TopicItem *titem = static_cast<TopicItem*>(item);
    QString url = titem->url;

    emit highlighted(url);
    topicListBox->viewport()->setCursor(url.isEmpty()?
                                        arrowCursor : pointingHandCursor);
}


void IndexWidget::onTopicViewport()
{
    topicListBox->viewport()->setCursor(arrowCursor);
    emit highlighted(QString());
}


void IndexWidget::readIndexFromFile(const QString &fileName)
{
    QFileInfo fi(fileName);
    QString name = fi.baseName();
    
    QFile f(fileName);
    if (!f.open(IO_ReadOnly)) {
        qDebug("Could not read doc index: %s", fileName.latin1());
        return;
    }

    QDomDocument dom;
    if (!dom.setContent(&f) || dom.doctype().name() != _("doxindex0")) {
        qDebug("Not a valid doxindex file: %s", fileName.latin1());
        return;
    }
    f.close();

    IndexFile *index = new IndexFile;
    indexFiles.append(index);

    QDomElement docEl = dom.documentElement();
    QDomElement titleEl = docEl.namedItem(_("title")).toElement();
    QDomElement baseEl = docEl.namedItem(_("base")).toElement();
    index->indexName = name;
    index->title = titleEl.firstChild().toText().data();
    index->base = baseEl.attribute(_("href"));
    if (!index->base.isEmpty())
        index->base += _("/");

    char fileno = indexFiles.count()-1;
    readEntryList(docEl, fileno);
}


void IndexWidget::readEntryList(const QDomElement &el, char fileno)
{
    QDomElement childEl = el.firstChild().toElement();
    while (!childEl.isNull()) {
        if (childEl.tagName() == _("item")) {
            QString primary = childEl.attribute(_("primary"));
            QString topic = childEl.attribute(_("topic"));
            QString url = childEl.attribute(_("url"));
            
            IndexItem *item = 0;
            QMap<QString,QListBoxItem*>::ConstIterator it = indexMap.find(primary);
            if (it != indexMap.end()) {
                item = static_cast<IndexItem*>(*it);
            } else {
                item = new IndexItem(indexListBox, primary);
                indexMap.insert(primary, item);
            }
            item->addTopic(fileno, topic, url);
        }

        childEl = childEl.nextSibling().toElement();
    }
}


void IndexWidget::showTerm(const QString &name)
{
    tabParent->showPage(this);

    termEdit->setText(name);
    if (indexListBox->currentText() == name)
        indexItemSelected(indexListBox->item(indexListBox->currentItem()));
}
