/***************************************************************************
 *   Copyright (C) 2001-2002 by Bernd Gehrmann                             *
 *   bernd@mail.berlios.de                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License, version 2       *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "richtextfactory.h"
#include "richtextview.h"
#include "global.h"
#include "../config.h"
#include <qapplication.h>
#include <qdir.h>
#include <qdragobject.h>
#include <qsettings.h>
#include <qurl.h>


RichtextFactory::RichtextFactory()
    : QMimeSourceFactory(),
      whatisRegExp(_("(\\S+)\\s+\\(([^)]+)\\)\\s*(\\[[^\\]]+\\])?\\s*-\\s+([^\\n]*)\\n"))
{
    QSettings *settings = Global::settings();
    int manCacheSize     = settings->readNumEntry(_("Dox/Cache/Man"),     2000);
    int infoCacheSize    = settings->readNumEntry(_("Dox/Cache/Info"),    2000);
    int whatisCacheSize  = settings->readNumEntry(_("Dox/Cache/Whatis"),  50);
    int aproposCacheSize = settings->readNumEntry(_("Dox/Cache/Apropos"), 50);
    int htdigCacheSize   = settings->readNumEntry(_("Dox/Cache/Htdig"),   50);
    int pydocCacheSize   = settings->readNumEntry(_("Dox/Cache/Pydoc"),   2000);
    int perldocCacheSize = settings->readNumEntry(_("Dox/Cache/Perldoc"), 100);
    int perlfaqCacheSize = settings->readNumEntry(_("Dox/Cache/Perlfaq"), 100);

    // cost is in units of QChar==16bit => 1kb = 500 units
    manCache.setMaxCost(manCacheSize*500);
    infoCache.setMaxCost(infoCacheSize*500);
    whatisCache.setMaxCost(whatisCacheSize*500);
    aproposCache.setMaxCost(aproposCacheSize*500);
    htdigCache.setMaxCost(htdigCacheSize*500);
    pydocCache.setMaxCost(pydocCacheSize*500);
    perldocCache.setMaxCost(perldocCacheSize*500);
    perlfaqCache.setMaxCost(perlfaqCacheSize*500);
    helpCache.setMaxCost(5000*500);
}


QString RichtextFactory::makeAbsolute(const QString &name, const QString &context) const
{
    qDebug("Make abs %s,%s", name.latin1(), context.latin1());
    if (name.startsWith(_("man:")))
        return makeAbsoluteUrl(name, "man");
    else if (name.startsWith(_("apropos:")))
        return makeAbsoluteUrl(name, "apropos");
    else if (name.startsWith(_("whatis:")))
        return makeAbsoluteUrl(name, "whatis");
    else if (name.startsWith(_("info:")))
        return makeAbsoluteUrl(name, "info");
    else if (name.startsWith(_("htdig:")))
        return makeAbsoluteUrl(name, "htdig");
    else if (name.startsWith(_("pydoc:")))
        return makeAbsoluteUrl(name, "pydoc");
    else if (name.startsWith(_("perlfaq:")))
        return makeAbsoluteUrl(name, "perlfaq");
    else if (name.startsWith(_("perldoc:")))
        return makeAbsoluteUrl(name, "perldoc");
    else if (name.startsWith(_("help:")))
        return makeAbsoluteUrl(name, "help");

    if (name.isNull() || name[0] != '/') {
        QString s = QFileInfo(context).dirPath() + _("/") + name;
        //        qDebug("abs: %s", s.local8Bit().data());
        return s;
    } else
        return name;
    //    return QMimeSourceFactory::makeAbsolute(name, context);
}


QString RichtextFactory::makeAbsoluteUrl(const QString &url, const char *protocol) const
{
    QString prot = _(protocol);
    QString name = url.mid(prot.length()+1);
    if (name.startsWith(_("//")))
        name.remove(0, 2);
    if (name.right(1) == _("/"))
        name.remove(name.length()-1, 1);
    
    //    qDebug("Return %s", (prot + _("://") + name).latin1());
    return prot + _("://") + name;
}


const QMimeSource *RichtextFactory::data(const QString &name) const
{
    //    qDebug("Trying '%s'", name.latin1());
    QString page = name;
    if (page.startsWith(_("man://"))) {
        page.remove(0, 6);
        // TODO: List of all manpages?
        if (page.isEmpty())
            return 0;
        QUrl::decode(page);
        return new QTextDrag(manPage(page));
    } else if (page.startsWith(_("whatis://"))) {
        page.remove(0, 9);
        if (page.isEmpty())
            return 0;
        page.replace(QRegExp(_("&")), _("?"));
        QUrl::decode(page);
        return new QTextDrag(aproposPage(page, false));
    } else if (page.startsWith(_("apropos://"))) {
        page.remove(0, 10);
        if (page.isEmpty())
            return 0;
        page.replace(QRegExp(_("&")), _("?"));
        QUrl::decode(page);
        return new QTextDrag(aproposPage(page, true));
    } else if (page.startsWith(_("info://"))) {
        page.remove(0, 7);
        if (page.isEmpty())
            return 0;
        QUrl::decode(page);
        return new QTextDrag(infoPage(page));
    } else if (page.startsWith(_("htdig://"))) {
        page.remove(0, 8);
        if (page.isEmpty())
            return 0;
        page.replace(QRegExp(_("&")), _("?"));
        QUrl::decode(page);
        return new QTextDrag(htdigPage(page));
    } else if (page.startsWith(_("pydoc://"))) {
        page.remove(0, 8);
        if (page.isEmpty())
            return 0;
        QUrl::decode(page);
        return new QTextDrag(pydocPage(page));
    } else if (page.startsWith(_("perldoc://"))) {
        page.remove(0, 10);
        if (page.isEmpty())
            return 0;
        QUrl::decode(page);
        return new QTextDrag(perldocPage(page));
    } else if (page.startsWith(_("perlfaq://"))) {
        page.remove(0, 10);
        if (page.isEmpty())
            return 0;
        QUrl::decode(page);
        return new QTextDrag(perlfaqPage(page));
    } else if (page.startsWith(_("help://"))) {
        page.remove(0, 7);
        if (page.isEmpty())
            return 0;
        QUrl::decode(page);
        return helpPage(page);
    }

    const QMimeSource *result = QMimeSourceFactory::data(name);
    if (!result) {
        return new QTextDrag(_("<html>File not found: %1</html>").arg(name));
    }
    return result;
}


QString RichtextFactory::pageFromCommand(const QCString &command) const
{
    qApp->processEvents();
    
    //    qDebug("Command: %s", command.data());
    
    // TODO: Replace with QProcess
    FILE *f = popen((command + " < /dev/null").data(), "r");
    if (!f)
        return QString::null;
    // Make sure the QTextStream is deleted before we pclose() the file.
    QTextStream *stream = new QTextStream(f, IO_ReadOnly);
    QString s = stream->read();
    delete stream;
    pclose(f);

    qApp->processEvents();

    return s;
}


QString RichtextFactory::manPage(const QString &name) const
{
    QString *item = manCache.find(name);
    if (item) {
        //        qDebug("found in cache");
        return (*item);
    }
    
    QCString command = "dox-man2html '";
    command += name.latin1();
    command += "'";

    QString s = pageFromCommand(command);
    if (s.isNull())
        return RichtextView::trUtf8("Could not start dox-man2html");

    manCache.insert(name, s);
    return s;
}


QString RichtextFactory::aproposPage(const QString &name, bool apropos) const
{
    StringCache &cache = apropos? aproposCache : whatisCache;
    QString *item = cache.find(name);
    if (item) {
        //        qDebug("found in cache");
        return (*item);
    }
    
    QCString command;
    QCString page = name.local8Bit();
    
    int pos;
    while ((pos = page.findRev('?')) != -1) {
        QCString query = page.mid(pos+1);
        page.truncate(pos);
        if (query == "method=regex")
            command.prepend("-r ");
        else if (query == "method=wildcard")
            command.prepend("-w ");
    }

    command.prepend(apropos? "apropos " : "whatis ");
    command += page;

    QString block = pageFromCommand(command);
    if (block.isNull())
        return RichtextView::trUtf8("Could not start apropos or whatis");

    block += _("\n");
    
    QString s = _("<html>");
    pos = 0;
    while (whatisRegExp.search(block, pos) != -1) {
        QString page = whatisRegExp.cap(1);
        QString section = whatisRegExp.cap(2);
        QString title = whatisRegExp.cap(4);
        QString pageEncoded = page;
        QUrl::encode(pageEncoded);
        s += _("<a href=\"man://%4.%1\">%2 - %3</a><br>")
            .arg(section).arg(page).arg(title).arg(pageEncoded);
        pos += whatisRegExp.matchedLength();
    }
    s += _("</html>");
    
    cache.insert(name, s);
    return s;
}


QString RichtextFactory::infoPage(const QString &name) const
{
    QString *item = infoCache.find(name);
    if (item) {
        //        qDebug("found in cache");
        return (*item);
    }
    
    QCString command = "dox-info2html '";
    command += name.latin1();
    command += "'";

    QString s = pageFromCommand(command);
    if (s.isNull())
        return RichtextView::trUtf8("Could not start dox-info2html");

    infoCache.insert(name, s);
    return s;
}


QString RichtextFactory::htdigPage(const QString &name) const
{
    QString *item = htdigCache.find(name);
    if (item) {
        //        qDebug("found in cache");
        return (*item);
    }

    QCString page = name.utf8();

    QCString op = "and";
    QCString sort = "title";
    int pos;
    while ((pos = page.findRev('?')) != -1) {
        QCString query = page.mid(pos+1);
        page.truncate(pos);
        if (query == "operator=or")
            op = "or";
        else if (query == "sort=date")
            sort = "date";
        else if (query == "sort=score")
            sort = "score";
    }

    QString words = QString::fromUtf8(page);
    QUrl::encode(words);
    
    QCString command = HTSEARCH;
    command += " -c ";
    command += QFile::encodeName(QDir::homeDirPath());
    command += "/.dox/htdig/htdig.conf '";
    command += "words=";
    command += words.local8Bit();
    command += "&method=";
    command += op;
    command += "&matchesperpage=50&format=builtin-short&sort=";
    command += sort;
    command += "'";
    
    QString s = pageFromCommand(command);
    if (s.isNull())
        return RichtextView::trUtf8("Could not start htsearch");

    s.replace(QRegExp(_("http://localhost/")), _("file:/"));
    s.replace(QRegExp(_("Content-type: text/html\n\n")), _(""));

    htdigCache.insert(name, s);
    return s;
}


QString RichtextFactory::pydocPage(const QString &name) const
{
    QString *item = pydocCache.find(name);
    if (item) {
        //        qDebug("found in cache");
        return (*item);
    }
    
    QCString command = "python ";
    command += QFile::encodeName(Global::resource(_("share/dox/scripts/dox_pydoc.py")));
    command += " -w ";
    command += name.latin1();
    
    QString s = pageFromCommand(command);
    if (s.isNull())
        return RichtextView::trUtf8("Could not start Python with the dox_pydoc.py script.");

    pydocCache.insert(name, s);
    return s;
}


QString RichtextFactory::perldocPage(const QString &name) const
{
    QString *item = perldocCache.find(name);
    if (item) {
        //        qDebug("found in cache");
        return (*item);
    }
    
    QCString command = "perldoc -t -f ";
    command += name.latin1();
    
    QString block = pageFromCommand(command);
    if (block.isNull())
        return RichtextView::trUtf8("Could not start the perldoc command.");

    QString s = _("<pre>");
    s += block;
    s += _("</pre>");
    perldocCache.insert(name, s);
    return s;
}


QString RichtextFactory::perlfaqPage(const QString &name) const
{
    QString *item = perlfaqCache.find(name);
    if (item) {
        //        qDebug("found in cache");
        return (*item);
    }
    
    QCString command = "( perldoc -u -q ";
    command += name.latin1();
    command += " | pod2html )";
    
    QString s = pageFromCommand(command);
    if (s.isNull())
        return RichtextView::trUtf8("Could not start the perldoc command.");

    perlfaqCache.insert(name, s);
    return s;
}


const QMimeSource *RichtextFactory::helpPage(const QString &name) const
{
    QString *item = helpCache.find(name);
    if (item) {
        //        qDebug("found in cache");
        return new QTextDrag(*item);
    }

    int pos = name.find('/');
    if (pos == -1)
        return new QTextDrag(RichtextView::trUtf8("Path not complete."));
    QString relPath = _("default/") + name.left(pos) + _("/");
    QString innerFileName = name.mid(pos+1);

    QString fileName;
    
    QStringList paths = QStringList::split(_(":"), _(KDEHTMLPATH));
    QStringList::ConstIterator pit;
    for (pit = paths.begin(); pit != paths.end(); ++pit) {
        QString candidate = (*pit) + relPath + innerFileName;
        if (QFile::exists(candidate)) {
            return QMimeSourceFactory::data(candidate);
            break;
        }
        candidate = (*pit) + relPath + _("index.cache.bz2");
        if (QFile::exists(candidate)) {
            fileName = candidate;
            break;
        }
    }
    
    if (fileName.isNull())
        return new QTextDrag(RichtextView::trUtf8("File not found."));

    QCString command = "bzip2 -d -c ";
    command += QFile::encodeName(fileName);
    
    QString s = pageFromCommand(command);
    if (s.isNull())
        return new QTextDrag(RichtextView::trUtf8("Could not start the bzip2 command on %1.").arg(command));
    
    int innerPos = s.find(_("<FILENAME filename=\"%1\">").arg(innerFileName));
    if (innerPos == -1)
        return new QTextDrag(RichtextView::trUtf8("Could not find file %1 inside cache file").arg(innerFileName));
    
    innerPos += 22 + innerFileName.length();
    int innerPos2 = s.find(_("</FILENAME>"), innerPos);
    if (innerPos2 == -1)
        innerPos2 = s.length();
    s = s.mid(innerPos, innerPos2-innerPos);
    
    helpCache.insert(name, s);
    return new QTextDrag(s);
}
