/***************************************************************************
 *   Copyright (C) 1999-2001 by Matthias Hoelzer-Kluepfel                  *
 *   hoelzer@kde.org                                                       *
 *   Copyright (C) 2001 by Bernd Gehrmann                                  *
 *   bernd@kdevelop.org                                                    *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License, version 2       *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include "htdig.h"
#include "../config.h"
#include <stdlib.h>
#include <unistd.h>
#include <qapplication.h>
#include <qdir.h>
#include <qdragobject.h>
#include <qfile.h>
#include <qlabel.h>
#include <qlayout.h>
#include <qmessagebox.h>
#include <qmime.h>
#include <qpixmap.h>
#include <qprocess.h>
#include <qprogressbar.h>
#include <qtextstream.h>
#include <qtimer.h>

#define _(s) QString::fromLatin1(s)


static QPixmap pixmap(const QString &name)
{
    const QMimeSource *mime = QMimeSourceFactory::defaultFactory()->data(name);
    QPixmap pix;
    QImageDrag::decode(mime, pix);
    return pix;
}


static QString resource(const QString &relpath)
{
    return _(PREFIX) + QString::fromLatin1("/") + relpath;
}


static QStringList resources(const QString &relpath)
{
    QStringList res;
    
    QDir d(QString::fromLocal8Bit(PREFIX) + _("/") + relpath);
    const QFileInfoList *l = d.entryInfoList();

    if (!l)
        return QStringList();
    
    QFileInfoListIterator it(*l);
    ++it; ++it;
    for (; it.current(); ++it) {
        QString name = it.current()->absFilePath();
        qDebug("File: %s", name.latin1());
        res.append(name);
    }

    return res;
}


HtdigDialog::HtdigDialog(QWidget *parent, const char *name)
    : QDialog(parent, name)
{
    setCaption(trUtf8("Generating search index"));

    proc = 0;
    QDir d = QDir::home();
    if (!d.exists(_(".dox")))
        d.mkdir(_(".dox"));
    if (!d.cd(_(".dox"))) {
        QMessageBox::information(this, trUtf8("dox-htdig"), trUtf8("Could not create .dox directory"));
        exit(1);
    }
    if (!d.exists(_("htdig")))
        d.mkdir(_("htdig"));
    if (!d.cd(_("htdig"))) {
        QMessageBox::information(this, trUtf8("dox-htdig"), trUtf8("Could not create .dox directory"));
        exit(1);
    }
    indexdir = d.absPath();
        
    QGridLayout *grid = new QGridLayout(this, 5, 3, 6);
    
    l1 = new QLabel(trUtf8("Scanning for files"), this);
    grid->addMultiCellWidget(l1, 0, 0, 1, 2);
    
    filesLabel = new QLabel(this);
    grid->addWidget(filesLabel, 1, 2);
    setFilesScanned(0);
    
    check1 = new QLabel(this);
    grid->addWidget(check1, 0, 0);
    
    l2 = new QLabel(trUtf8("Extracting search terms"), this);
    grid->addMultiCellWidget(l2, 2, 2, 1,2);
    
    bar = new QProgressBar(this);
    grid->addWidget(bar, 3,2);
    
    check2 = new QLabel(this);
    grid->addWidget(check2, 2,0);
    
    l3 = new QLabel(trUtf8("Generating index..."), this);
    grid->addMultiCellWidget(l3, 4, 4, 1,2);
    
    check3 = new QLabel(this);
    grid->addWidget(check3, 4, 0);

    filesScanned = 0;
    filesDigged = 0;

    setState(0);
    
    setMinimumWidth(300);
}


HtdigDialog::~HtdigDialog()
{}


void HtdigDialog::done(int r)
{
    if (!r)
        proc->kill();
    QDialog::done(r);
}


void HtdigDialog::setFilesScanned(int n)
{
    filesLabel->setText(trUtf8("Files to index: %1").arg(n));
}


void HtdigDialog::setFilesToDig(int n)
{
    bar->setTotalSteps(n);
}


void HtdigDialog::setFilesDigged(int n)
{
    bar->setProgress(n);
}


void HtdigDialog::setState(int n)
{
    QPixmap unchecked = pixmap(_("unchecked"));
    QPixmap checked = pixmap(_("checked"));
    
    check1->setPixmap( n > 0 ? checked : unchecked);
    QFont font1(l1->font());
    font1.setBold(n == 0);
    l1->setFont(font1);
    check2->setPixmap( n > 1 ? checked : unchecked);  
    QFont font2(l2->font());
    font2.setBold(n == 1);
    l2->setFont(font2);
    check3->setPixmap( n > 2 ? checked : unchecked);  
    QFont font3(l3->font());
    font3.setBold(n == 2);
    l3->setFont(font3);

    qApp->processEvents();
    QApplication::syncX();
}


void HtdigDialog::addCandidate(QString url)
{
    int pos = url.findRev(_("#"));
    if (pos != -1)
        url.truncate(pos);
    if ((url.startsWith(_("/")) || url.startsWith(_("file://")))
        && !files.contains(url)) {
        //        qDebug("Adding %s", url.local8Bit().data());
        files.append(url);
        setFilesScanned(++filesScanned);
    }
}


void HtdigDialog::parseTocDom(const QString &base, QDomElement &el, int level)
{
    QString tag = _("tocsect%1").arg(level);
    QDomElement childEl = el.firstChild().toElement();
    while (!childEl.isNull()) {
        if (childEl.tagName() == tag) {
            QString url = childEl.attribute(_("url"));
            addCandidate(url.isEmpty()? url : (base+url));
            parseTocDom(base, childEl, level+1);
        }
        childEl = childEl.nextSibling().toElement();
    }
}


void HtdigDialog::parseTocFiles()
{
    qDebug("Parsing toc files");
    
    QStringList tocs = resources(_("share/dox/tocs"));

    QStringList::ConstIterator it;
    for (it = tocs.begin(); it != tocs.end(); ++it) {
        QFile f(*it);
        if (!f.open(IO_ReadOnly)) {
            qDebug("Could not read doc toc: %s", (*it).local8Bit().data());
            continue;
        }
        QDomDocument dom;
        if (!dom.setContent(&f) || dom.doctype().name() != _("doxtoc0")) {
            qDebug("Not a valid doxtoc file: %s", (*it).local8Bit().data());
            continue;
        }
        f.close();

        QDomElement docEl = dom.documentElement();
        QString base;
        QDomElement childEl = docEl.firstChild().toElement();
        while (!childEl.isNull()) {
            if (childEl.tagName() == _("tocsect1")) {
                QString name = childEl.attribute(_("name"));
                QString url = childEl.attribute(_("url"));
                addCandidate(url.isEmpty()? url : (base+url));
                parseTocDom(base, childEl, 2);
            } else if (childEl.tagName() == _("base")) {
                base = childEl.attribute(_("href"));
                if (!base.isEmpty())
                    base += _("/");
            }
            childEl = childEl.nextSibling().toElement();
        }

        qApp->processEvents();
    }
}


void HtdigDialog::parseIndexFiles()
{
    qDebug("Parsing index files");
    
    QStringList indices = resources(_("share/dox/indices"));

    QStringList::ConstIterator it;
    for (it = indices.begin(); it != indices.end(); ++it) {
        QFile f(*it);
        if (!f.open(IO_ReadOnly)) {
            qDebug("Could not read index: %s", (*it).local8Bit().data());
            continue;
        }
        QDomDocument dom;
        if (!dom.setContent(&f) || dom.doctype().name() != _("doxindex0")) {
            qDebug("Not a valid doxindex file: %s", (*it).local8Bit().data());
            continue;
        }
        f.close();

        QDomElement docEl = dom.documentElement();
        QString base;
        QDomElement childEl = docEl.firstChild().toElement();
        while (!childEl.isNull()) {
            if (childEl.tagName() == _("item")) {
                QString url = childEl.attribute(_("url"));
                addCandidate(url.isEmpty()? url : (base+url));
            } else if (childEl.tagName() == _("base")) {
                base = childEl.attribute(_("href"));
                if (!base.isEmpty())
                    base += _("/");
            }
            childEl = childEl.nextSibling().toElement();
        }

        qApp->processEvents();
    }
}


bool HtdigDialog::createConfig()
{
    QString wrapper = resource(_("share/dox/htdig/wrapper.html"));
    if (wrapper.isEmpty())
        return false;
    wrapper = wrapper.left(wrapper.length()-12);
    
    QFile f(indexdir + _("/htdig.conf"));
    if (f.open(IO_WriteOnly)) {
        QTextStream ts(&f);
        
        ts << "database_dir:\t\t" << indexdir << endl;
        ts << "start_url:\t\t`" << indexdir << "/files`" << endl;
        ts << "local_urls:\t\thttp://localhost/=/" << endl;
        ts << "local_urls_only:\ttrue" << endl;
        ts << "maximum_pages:\t\t1" << endl;
        ts << "image_url_prefix:\t\t" << wrapper << endl;
        ts << "star_image:\t\t" << wrapper << "star.png" << endl;
        ts << "star_blank:\t\t" << wrapper << "star_blank.png" << endl;
        ts << "compression_level:\t\t6" << endl;
        ts << "max_hop_count:\t\t0" << endl;
        
        ts << "search_results_wrapper:\t" << wrapper << "wrapper.html" << endl;
        ts << "nothing_found_file:\t" << wrapper << "nomatch.html" << endl;
        ts << "syntax_error_file:\t" << wrapper << "syntax.html" << endl;
        ts << "bad_word_list:\t\t" << wrapper << "bad_words" << endl;
        
        f.close();
        return true;
    }
    
    return false;
}


#define CHUNK_SIZE 100


bool HtdigDialog::generateIndex()
{
    setState(1);

    // run htdig

    bool initial = true;
    bool done = false;
    int  count = 0;
    
    filesToDig = files.count();
    setFilesToDig(filesToDig);
    
    //    QDir d; d.mkdir(indexdir);
    
    while (!done) {
        // kill old process
        delete proc;
        
        // prepare new process
        proc = new QProcess();
        proc->addArgument(_(HTDIG));
        proc->addArgument(_("-c"));
        proc->addArgument(indexdir + _("/htdig.conf"));
        if (initial) {
            proc->addArgument(_("-i"));
            initial = false;
        }
        
        qDebug("Running htdig");
        
        connect(proc, SIGNAL(processExited()),
                this, SLOT(htdigExited()));
        
        htdigRunning = true;
        
        // write out file
        QFile f(indexdir + _("/files"));
        if (!f.open(IO_WriteOnly)) {
            qDebug("Could not open `files` for writing");
            return false;
	}
        
        QTextStream ts(&f);
        for (int i=0; i<CHUNK_SIZE; ++i, ++count) {
            if (count >= filesToDig) {
                done = true;
                break;
            }
            ts << _("http://localhost/") + files[count] << endl;
        }
        f.close();
        
        // execute htdig
        proc->start();
        while (htdigRunning && proc->isRunning())
            qApp->processEvents();
        
        if (!proc->normalExit() || proc->exitStatus() != 0) {
            QMessageBox::information(this, trUtf8("dox-htdig"), trUtf8("Running htdig failed"));
            delete proc;
            return false;
        }
        
        filesDigged += CHUNK_SIZE;
        setFilesDigged(QMIN(filesToDig, filesDigged));
        qApp->processEvents();
    }

    setState(2);

    // run htmerge -----------------------------------------------------
    delete proc;
    proc = new QProcess();
    proc->addArgument(_(HTMERGE));
    proc->addArgument(_("-c"));
    proc->addArgument(indexdir + _("/htdig.conf"));

    qDebug("Running htmerge");
    
    connect(proc, SIGNAL(processExited()),
            this, SLOT(htmergeExited()));
    
    htmergeRunning = true;
    
    proc->start();
    
    while (htmergeRunning && proc->isRunning())
        qApp->processEvents();
    
    if (!proc->normalExit() || proc->exitStatus() != 0) {
        delete proc;
        return false;
    }

    delete proc;
    
    setState(3);
    
    return true;
}


void HtdigDialog::htdigExited()
{
    qDebug("htdig terminated");
    htdigRunning = false;
}


void HtdigDialog::htmergeExited()
{
    qDebug("htmerge terminated");
    htmergeRunning = false;
}


int main(int argc, char **argv)
{
    QApplication app(argc, argv);

    HtdigDialog *search = new HtdigDialog(0, "progress dialog");
    search->show();
    qApp->processEvents();
    QApplication::syncX();

    // Assemble url list
    search->parseTocFiles();
    search->parseIndexFiles();

    // Write htdig.conf file
    if (!search->createConfig())
        return 1;

    // Do it :-)
    search->generateIndex();

    sleep(1);
    
    delete search;

    // Make sure the main event loop runs once
    QTimer::singleShot(0, &app, SLOT(quit()));
    return app.exec();
}
