/***************************************************************************
 *   Copyright (C) 2001 by Bernd Gehrmann                                  *
 *   bernd@mail.berlios.de                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License, version 2       *
 *   as published by the Free Software Foundation.                         *
 *                                                                         *
 ***************************************************************************/

#include <qdom.h>
#include <qfile.h>
#include <qregexp.h>
#include <qtextstream.h>
#include <stdio.h>

#define _(s) QString::fromLatin1(s)


void quote(QString *s)
{
    s->replace(QRegExp(_("<")), _("&lt;"));
    s->replace(QRegExp(_(">")), _("&gt;"));
}


int main(int argc, char **argv)
{
    if (argc != 5) {
        fprintf(stderr, "Usage: dox-tags2index inputfile.tags baseurl outputfile.index outputfile.toc\n");
        return 1;
    }

    QString tagsFileName = QFile::decodeName(argv[1]);
    QString baseUrl = QFile::decodeName(argv[2]);
    QString indexFileName = QFile::decodeName(argv[3]);
    QString tocFileName = QFile::decodeName(argv[4]);

    int pos = tagsFileName.findRev(_("."));
    if (pos == -1)
        pos = tagsFileName.length();
    QString title = tagsFileName.left(pos);
    
    QFile tagsf(tagsFileName);
    if (!tagsf.open(IO_ReadOnly)) {
        fprintf(stderr, "Could not read tag file: %s\n", tagsFileName.local8Bit().data());
        return 1;
    }
    QDomDocument doc;
    if (!doc.setContent(&tagsf) || doc.documentElement().nodeName() != _("tagfile")) {
        fprintf(stderr, "Not a valid doxygen tags file: %s\n", tagsFileName.local8Bit().data());
        return 1;
    }
    tagsf.close();

    //
    // index
    //

    QFile indexf(indexFileName);
    if (!indexf.open(IO_WriteOnly)) {
        fprintf(stderr, "Could not write index file: %s\n", indexFileName.local8Bit().data());
        return 1;
    }
    QTextStream *indexstream = new QTextStream(&indexf);
    indexstream->setEncoding(QTextStream::UnicodeUTF8);
        
    (*indexstream) << "<!DOCTYPE doxindex0>" << endl;
    (*indexstream) << "<doxindex0>" << endl;
    (*indexstream) << "<title>" << title << "</title>" << endl;
    (*indexstream) << "<base href=\"" << baseUrl << "\"/>" << endl;

    QDomElement el = doc.documentElement();

    QValueList<QCString> topicList;
    bool hasNamespaces = false;

    QDomElement childEl = el.firstChild().toElement();
    while (!childEl.isNull()) {
        if (childEl.tagName() == _("compound")) {

            QString name = childEl.namedItem(_("name")).firstChild().toText().data();
            QString filename = childEl.namedItem(_("filename")).firstChild().toText().data();
            QString kind = childEl.attribute(_("kind"));

            hasNamespaces |= (kind == _("namespace"));
            
            quote(&name);
            
            QCString primary = name.utf8();
            QCString topic = name.utf8();
            QCString url = filename.utf8();
            (*indexstream)
                << "<item primary=\"" << primary
                << "\" topic=\"" << topic
                << "\" url=\"" << url
                << "\"/>" << endl;
            
            QDomElement grandchildEl = childEl.firstChild().toElement();
            while (!grandchildEl.isNull()) {
                if (grandchildEl.tagName() == _("member")) {
                    QString membername = grandchildEl.namedItem(_("name")).firstChild().toText().data();
                    QString anchor = grandchildEl.namedItem(_("anchor")).firstChild().toText().data();
                    QString arglist = grandchildEl.namedItem(_("arglist")).firstChild().toText().data();

                    quote(&membername);
                    
                    // leave out constructors and destructors
                    if (membername != name && membername != (_("~") + name)) {
                        primary = (membername + _("()")).utf8();
                        topic = (membername + _("(), ") + name + _(" ") + kind).utf8();
                        url = (filename + _("#") + anchor).utf8();
                        if (!topicList.contains(topic)) {
                            (*indexstream)
                                << "<item primary=\"" << primary
                                << "\" topic=\"" << topic
                                << "\" url=\"" << url
                                << "\"/>" << endl;
                            topicList.append(topic);
                        }
                    }
                }
                
                grandchildEl = grandchildEl.nextSibling().toElement();
            }
            
        }
        
        childEl = childEl.nextSibling().toElement();
            
    }

    (*indexstream) << "</doxindex0>" << endl;

    delete indexstream;
    indexf.close();

    //
    // toc
    //

    QFile tocf(tocFileName);
    if (!tocf.open(IO_WriteOnly)) {
        fprintf(stderr, "Could not write toc file: %s\n", tocFileName.local8Bit().data());
        return 1;
    }
    QTextStream *tocstream = new QTextStream(&tocf);
    tocstream->setEncoding(QTextStream::UnicodeUTF8);

    (*tocstream) << "<!DOCTYPE doxtoc0>" << endl;
    (*tocstream) << "<doxtoc0>" << endl;
    (*tocstream) << "<title>" << title << "</title>" << endl;
    (*tocstream) << "<base href=\"" << baseUrl << "\"/>" << endl;
    (*tocstream) << "<tocsect1 name=\"Main page\" url=\"index.html\"/>" << endl;
    if (hasNamespaces) {
        (*tocstream) << "<tocsect1 name=\"Namespace list\" url=\"namespaces.html\">" << endl;

        QDomElement childEl = el.firstChild().toElement();
        while (!childEl.isNull()) {
            if (childEl.tagName() == _("compound") && childEl.attribute(_("kind")) == "namespace") {
                QString name = childEl.namedItem(_("name")).firstChild().toText().data();
                QString filename = childEl.namedItem(_("filename")).firstChild().toText().data();
            
                quote(&name);
                
                QCString title = name.utf8();
                QCString url = filename.utf8();
                (*tocstream) << "  <tocsect2 name=\"" << title << "\" url=\"" << url << "\"/>" << endl;
            }
            childEl = childEl.nextSibling().toElement();
        }
            
        (*tocstream) << "</tocsect1>" << endl;
    }        

    // Block just for structuring purposes
    {        
        (*tocstream) << "<tocsect1 name=\"Compound List\" url=\"annotated.html\">" << endl;
        
        QDomElement childEl = el.firstChild().toElement();
        while (!childEl.isNull()) {
            if (childEl.tagName() == _("compound") && childEl.attribute(_("kind")) != "namespace") {
                QString name = childEl.namedItem(_("name")).firstChild().toText().data();
                QString filename = childEl.namedItem(_("filename")).firstChild().toText().data();
            
                quote(&name);
                
                QCString title = name.utf8();
                QCString url = filename.utf8();
                (*tocstream) << "  <tocsect2 name=\"" << title << "\" url=\"" << url << "\"/>" << endl;
            }
            childEl = childEl.nextSibling().toElement();
        }

        (*tocstream) << "</tocsect1>" << endl;
    }
    
    (*tocstream) << "<tocsect1 name=\"Class Hierarchy\" url=\"hierarchy.html\"/>" << endl;
    (*tocstream) << "<tocsect1 name=\"File List\" url=\"files.html\"/>" << endl;
    if (hasNamespaces)
        (*tocstream) << "<tocsect1 name=\"Namespace Members\" url=\"namespacemembers.html\"/>" << endl;
    (*tocstream) << "<tocsect1 name=\"Compound Members\" url=\"functions.html\"/>" << endl;

    (*tocstream) << "</doxtoc0>" << endl;

    delete tocstream;
    tocf.close();
}
